// ---------------------------------------------------------------
//    consol.c
//
//     このプログラムは最初 Linux で開発したソースを WINDOWS　に最低限の手間で動作するよう
//     移植したものです。
//     移植のため、CYGWINのみをインストールしました。 
//     他のコンパイラや WIN32API は一切使用していません。
//     CYGWINパッケージはサイズも大きく必要なければ使いたくないのですが
//     POSIX互換のGCC,PTHREAD,名前付FIFO　などがそのまま使えます。
//      (但し　これ以上ソースの修正、改良が必要なければ、実行時最低環境として2種のDLLを
//       所定の場所へCOPY しておくだけで実行できます。）
//
//     このプログラムは PIC24FJ64GA002 にあらかじめ書き込まれたブートローダと連携し、
//     ユーザが開発したPIC-AP（HEXファイル）をPIC書き込み器なしでプログラミングする
//     ホスト側の機能部分です。
//
//     動作環境 （WINDOWS-XP　で動作確認） 
//      1. RS232 が有効で COM1(=/dev/ttyS0) が定義されていること
//         通信速度等は (brt= 19200 8N1 cts/rts使用)であること。（以下にその設定方法を示す。)
//
//      2.動作環境として CIGWIN をインストールしてある場合　（ソースの修正ができる。）
//        (1) cygwin を起動し 以下cygwin tarminal画面より操作する。
//        (2) 適宜なサブディレクトリ下で consol.c を make する。    [>make]
//        (3) 通信速度等の設定のため、 mode.sh を１回起動する。     [>./mode.sh]
//        (4) (2)で作成した consol.exe を実行する。                 [>./consol]
//
//      3.動作環境として CIGWIN をインストールしてない場合
//        (1) 最低限の実行環境のため添付の "cygwin1.dll" , "cyggcc_s-1.dll"を c:\WINDOWS\ にコピーする。
//        (2) term.bat を起動し 以下 term画面(実体はコマンドプロンプト画面)より操作する。
//        (3) 通信速度等の設定のため、 mode.bat を１回起動する。    [>mode.bat]
//        (4) あらかじめ作成した consol.exe を実行する。            [>consol]
//
//     全体機能概要としては
//        (1) KBコマンド を PIC に送りPIC側で解釈され、応答を得る非同期ダム端末の機能
//            この機能だけみると、一般の通信ソフト（例えばTera Term) でも良いが         
//        (2) PIC側と連携し送受信するプロトコル部分（ブートローダ機能）がメイン
//	          writepic [hexfile path]        PGROM Write(0x4000-A800)
//
//     プログラムの概要 
//       Main スレッド  以下の２スレッドを生成  
//         (1) read_line()スレッド: 専らRS232(PIC)からの受信を行う。　
//         (2) kb_read()  スレッド: 専らKBからの入力,コマンド解析・実行を行う。
//        と協調し、非同期でトランザクションを処理を行う。（各入力でブロックさせないため）
//        ２スレッド間の情報転送は名前付きパイプで行っている。
//
//     プログラム上の特記事項
//         (1) HEX file を読み込んで  mallocで確保したメモリ上のimage に格納する.
//             併せて何かの参照用に"hexwork.txt"というファイルにimage を自動作成する。
//         (2) malloc buff(image) 有効セクタ数分(1セクタ=256byte) PIC に送る。
//         (3) PIC側の書き込みは1セクタ単位に実施され、完了の同期として"RDY" をやり取りするプロトコル
//             となっている。
//
//     LinuxソースからCygWinソース 変換に当たっての参考事項
//         (1) FIFO（名前付きパイプ）のオープンが Linux では "O_RDWR" で１個で済むが
//             CygWin では"O_RDONLY", "O_WRONLY" と別々にオープンしないと途中ブロックする。
//             これを解決するまで、ずいぶん時間がかかった。
//         (2) FIFO（名前付きパイプ）の作成場所
//             Linux では "/tmp/cdchost" を作るが CygWin をインストールしていない場合を考慮し
//             場所としてWINDOWS上の "c:\temp\cdchost"(=/cygdrive/c/temp/cdchost) を作った。
//             ファイル名(cdchost)はもともと識別のためで何でも良い。
//         (3) (2)で作ったファイル属性はWINDOWS上のブラウザで見るとショートカット属性になって
//             いるが心配ない。 CygWin で見ると "P"(pipeデバイス)になっている。
//         (4) RS232ポート(COM1)をOPENする時のデバイス名は"/dev/ttyS0" とLinuxと同じである。
//              CygWin で "ls -al /dev/*" と確認すると当該名が無いのでとまどうが心配ない。
//              試しに "ls -al /dev/ttyS0" と確認すると当該名が表示される。これは CygWinの
//              完成度の低さと思われる。
//         (5) LinuxではRS232ポートの通信速度等をプログラム(termios)で設定していたが、
//             コンパイルエラーにはならないが機能しないので削除した。代わりに手動で昔の
//             "mode.com" を使いシェル又はバッチファイルで設定するようにした。
//         (6) スレッドの終結時処理がLinux と微妙に異なるため若干追加した。
//              (pthread_cancel(),  pthread_testcancel() )
// ---------------------------------------------------------------
    #include <string.h>
    #include <stdio.h>
    #include <stdlib.h>
//  #include <termios.h>    // tcflush() 
    #include <unistd.h>		// read write
    #include <fcntl.h>		// open 
    #include <sys/types.h>	// open 
    #include <sys/stat.h>	//  
    #include <pthread.h>	// スレッド関連
    #include <errno.h>		// 標準　Error定義

typedef unsigned char   uchar;
typedef unsigned short int    word;	// 2byte
typedef unsigned int    uint;		// 4byte
typedef unsigned long   ulong;		// 8byte

// 重要な内部定義関数
void* kb_read( ) ;					// kb_read() スレッド:専らKBからの入力,コマンド解析・実行を行う。
void* ans_read_line( ) ;			// ans_read_line()スレッド:専らRS232(PIC)からの受信を行う。
short cmd_flash_write(char* path ); // PGROM Write AP本体
void  pipe_read_line(void) ;		// 専らFIFOからの受信を行う。

// HEX FILE関連　内部定義関数
short	read_hexfile(char *path) ;
short   write_hexdata(void ) ;		// hex file用
short	read_ihex(void) ;
short	parse_ihex(void ) ;
void	image_write( void ) ;		//  malloc buf 書き込み 1line分
void 	read_hex_string(void) ;
short 	getln(char *buf, FILE *fp) ;
short 	read_hex_byte(char *s) ;
word 	atohex(char *buf ) ;

// 共通内部定義関数
void delayms(short T) ;			// msec単位
void delayus(short T) ;			// usec単位
void delayns(short T) ;			// nsec単位

// local 変数定義
static  int  rs232_fd ;			// USB /dev/tty/ACM0 -> RS232 /dev/ttyS0(=COM1)
static  int  pipe_rd_fd ;		// pipe fd  for read
static  int  pipe_wr_fd ;		// pipe fd  for write
static  int  pipe_req ;			// pipe write要求 
static  char linebuf[80];		// 1 line buff (文字列)
static  char pipebuf[82] ;

//  hex Read 関連 common変数 
static uchar  databuf[40];		// 1line -> 2文字単位にbinary化
static uchar  *mptr;			// malloc buff pointer
static uint   CONFIG1, CONFIG2;		// readout config1,2
static int    Sector;			// 有効セクタ(256byte)数


//-----------------------------------------------------------------------------------
//  Main() 
//      Main スレッド  以下の２スレッドを生成  
//         (1) read_line()スレッド: 　専らRS232(PIC)からの受信を行う。　
//         (2) kb_read() スレッド:    専らKBからの入力,コマンド解析・実行を行う。
//      と協調し、非同期でトランザクションを処理を行う。（各入力でブロックさせないため）
//      ２スレッド間の情報転送は名前付きパイプで行っている。
//------------------------------------------------------------------------------------
int  main() {
	short r;
	pthread_t t1, t2 ;
/*
 	struct termios temp_term;	// tcflag_t c_iflag;      // input modes 
              				// tcflag_t c_oflag;      // output modes 
              				// tcflag_t c_cflag;      // control modes
              				// tcflag_t c_lflag;      // local modes
              				// cc_t     c_cc[NCCS];   // control chars
*/

// /dev/ttyS0 をオープンする。
	rs232_fd = open("/dev/ttyS0", O_RDWR  );
	if (rs232_fd < 0) {printf("PIC Open Error  \n"); return 0; }
	printf("PIC Open fd=%d OK \n", rs232_fd); 

/* -----  cwgwin でエラーにはならないが機能しないので　手動でmode.shを動かすこと。 -------------
	// 現在の端末設定を取得し、おまじないの設定(minicomと同等)を行う。これをしないと変な動作をする。 
	errno = 0;
  	if(tcgetattr(rs232_fd, &temp_term) == -1){perror("tcgetattr failure"); return 0; }

//	printf("iflag=%x oflag=%x cflag=%x lflag=%x \n",temp_term.c_iflag, temp_term.c_oflag,temp_term.c_cflag,temp_term.c_lflag);

	temp_term.c_iflag = 0x01;		// IGNBRK
	temp_term.c_oflag = 0x00;
	//temp_term.c_cflag = 0x08bf;		// brt= 38400 8N1 CREAD CLOCAL
	temp_term.c_cflag = B19200 | CS8 | CLOCAL | CREAD ;	// brt= 19200 8N1 CREAD CLOCAL
	temp_term.c_lflag = 0x00;

  	// 即時反映させる。
  	errno = 0;
  	if(tcsetattr(rs232_fd, TCSANOW, &temp_term) == -1)
		{perror("tcsetattr(temp_term) failure");return 0;}
	// 念のためFLASH する。　（ＩＯ未完了のストリームを破棄する）
  	tcflush(rs232_fd, TCIOFLUSH);  // termios  FLASH
-------------------------------------------------------------------------------------------------------------------
*/
// FIFO（名前付きパイプ）をオープンする
//     Linux では "/tmp/" に作るが WINDOWS　では"c:\temp\"(=/cygdrive/c/temp/) に作る。
//	pipe_wr_fd = open("/tmp/cdchost", O_WRONLY) ;
	pipe_wr_fd = open("/cygdrive/c/temp/cdchost", O_WRONLY) ;
	pipe_rd_fd = open("/cygdrive/c/temp/cdchost", O_RDONLY) ;
	if (pipe_wr_fd < 0) { 				// Error
	    r = mkfifo("/cygdrive/c/temp/cdchost", 0666 ) ;	// FIFO（名前付きパイプ）を新規生成
	    if (r < 0) {perror("mkfifo Error  \n"); return 0; }

	    pipe_wr_fd = open("/cygdrive/c/temp/cdchost", O_WRONLY) ;	// 再オープン
	    pipe_rd_fd = open("/cygdrive/c/temp/cdchost", O_RDONLY) ;	// 再オープン

	}
	printf("named_pipe c:\\temp\\cdchost Open OK \n");
	printf("usage: please Keyin [help], [q](=quit) \n");

//  スレッドの協調に必要なフラグのクリア
	pipe_req = 0 ;		// RS232 からの受信をパイプ経由で kb_read()スレッドに渡す要求有りを示す。

// スレッドの生成 kb_read()
	pthread_create( &t1, NULL, kb_read ,       (void *)NULL );
// スレッドの生成 ans_read_line()
	pthread_create( &t2, NULL, ans_read_line , (void *)NULL );

//  メイン処理開始   とりわけて何もしない
//	fprintf(stderr,":") ;				// プロンプト
	write(rs232_fd, "\r", 1) ; 			// dummy PIC に"\r"を送る

// kb_read()スレッドの終了を待つ(KB から QUIT入力)
	pthread_join( t1, NULL );	//  KB から QUIT入力
	pthread_cancel( t2 );		//  ans_read_line()スレッド をCANCEL
//	pthread_join( t2, NULL );	//  ans_read_line()スレッドの終了を待つ
	close(rs232_fd);
	close(pipe_wr_fd);
	close(pipe_rd_fd);
	return 0;			// 自分が親スレッドなので、終えると全スレッドが開放される。
}

//--------------------------------------------------
//  RS232(PIC) Read_line    スレッド
//     RS232(PIC) から１行受信し、そのまま表示する。
//--------------------------------------------------
void* ans_read_line( ) {
	short s, index ;
	short first_flg = 0 ;
	char wk[2], buf[82] ;

//    １文字毎にREADし１行単位に受信バッファに格納　最後は(\r)が入ってる。
	index = 0 ;
	while(1) {
	    s = read(rs232_fd,wk,1);
		pthread_testcancel();
	    if (s == -1)  { printf("PIC Read Error \n");  return NULL;}
	    buf[index++] = wk[0]; 
	    if (index > 80) index = 80;					// for buf overflow
	    if ( wk[0] == '\r') {						// line end
			buf[index++] = '\n'; buf[index] = '\0';		// \n add & buff stopper 
			if (pipe_req == 1) {
				s = write(pipe_wr_fd,buf,strlen(buf));			// BUG?  ブロックする。
	        	if (s == -1)  { printf("FIFO Write Error \n");  return NULL;}
			}
			if (first_flg == 0) first_flg = 1;
			else printf( buf);
			index = 0 ;			// 
 	    }
	}
	return NULL ;
}

//--------------------------------------------------
//  PIPE(RS232) Read_line   
//     PIPE(RS232) から１行受信する。
//--------------------------------------------------
void pipe_read_line( ) {
	short index ;
	char wk[2] ;

//    １文字毎にREADし１行単位に受信バッファに格納　最後は(\n)が入ってる。
	index = 0 ;
	while(1) {
	    read(pipe_rd_fd,wk,1);
		pthread_testcancel();

	    pipebuf[index++] = wk[0]; 
	    if (index > 80) index = 80;		// for buf overflow
	    if ( wk[0] == '\n') {			// line end
			pipebuf[index] = '\0';		// buff stopper 
			break ;	
	    }
	}

	return ;
}

//-------------------------------------------------------------------
//  KB Read スレッド
//      "q" で Break   pthread_exit( NULL );
//-------------------------------------------------------------------
void*  kb_read( ) {  	 
	char kb_buf[80];		// 
	char cmd[10];			// 入力解析　コマンド文字列
	char path[30];			// 入力解析　第２パラメータ　PATH文字列
	short  r ;
	while (1)  {				// 1行入力 --> 解析
		cmd[0] = '\0';
		fgets(kb_buf,72,stdin) ;    	// 1行入力 
		sscanf( kb_buf, "%s %s",cmd, path);

		if      (strcmp(cmd, "writepic") == 0)  {   // PGROM Write
			r = cmd_flash_write(path);
			if (r == 9) { ;	}		// Error でもここでは何もしない。 PIC にも送らない。
			}
		else if (strcmp(cmd, "q")        == 0)  {
			break;
			}
		else  {	
			strcat(kb_buf,"\r");		// kb_bufには \r が無いため。	
			write(rs232_fd, kb_buf, strlen(kb_buf) ) ; }     // AUX --> PIC に kb_buf そのまま送る
	}
	return NULL;
}

//------------------------------------------
//  cmd_flash_write()　　Command "writepic PATH " 
//      return 9:break(Error)  0:ok
//------------------------------------------
short cmd_flash_write(char* path) {		// PGROM Write
	short s;


//	HEX file を読み込んで メモリ上のimage に格納する.
//       & malloc buff(image) 有効セクタ数分　hexwork.txt　に書き出す
	s = read_hexfile( path ) ;
	if (s == 9 ) return 9;

	s = write_hexdata( );		// malloc buff(image) 有効セクタ数分 ターゲット PIC に送る
	if (s == 9 ) return 9;

	return 0 ;
}

// -----------------------------------------------------------------
//	malloc buff(image) 有効セクタ数分 ターゲット PIC に送る
// -----------------------------------------------------------------
short	write_hexdata( ) {

	char txbuf[80] ;
	word addr, d[8], *ptr ;
	short  i, j, k ;

//  PIC へのコマンドとして　key wordを送る。
	sprintf(txbuf, "writepic %d \r", Sector );	// KEY command
	write(rs232_fd, txbuf, strlen(txbuf) ) ;        // error: r = -1
	printf( txbuf ) ; printf( "\n" ) ;		// monitor

	// PIC からのRDY（準備完了）を待つ
	//      スレッド間インタフェースとして名前付きパイプが定義され
	//      このパイプを通じて、RS232 (PIC) からの応答を得る。
	pipe_req = 1 ;
	pipe_read_line() ;				// pipebuf: RDY\r\n
	pipe_req = 0 ;

printf("NEXT %s",pipebuf);	
	// RDY  -> next

//  malloc buff(0100-01ff)　２セクタ　AIVT(代替ベクタテーブル）
	addr = 0x0100; 
	ptr = (word*)mptr + 0x0100 ;
	for (i = 0 ; i < 2 ; i++ )  {
	    for (j = 0 ; j < 16 ; j++ )  {		// 8W * 16 = 128W
	    	for (k = 0 ; k < 8 ; k++ )  {		// 8W -> d[8] copy
		    d[k] = *ptr++;
		}
	    	sprintf(txbuf, "%04X  %04X %04X %04X %04X %04X %04X %04X %04X \r", addr,
			d[0],d[1],d[2],d[3],d[4],d[5],d[6],d[7] );
		write(rs232_fd, txbuf, strlen(txbuf) ) ;        // error: r = -1
		printf( txbuf ) ;  printf( "\n" ) ;
		addr +=  8 ;
		//delayms( 10) ;		// PIC側が遅いのでゆっくり送るため
	    }
	    //printf( "\n" ) ;

	    // PIC からのRDY（sector write）を待つ
	    pipe_req = 1 ;
	    pipe_read_line() ;				// pipebuf: RDY\r\n
	    pipe_req = 0 ;
	    if (strncmp(pipebuf, "RDY", 3) != 0 )  return 9;
	    // RDY  -> next
	}

//  malloc buff(4000-A800)  セクタ数分繰り返す。　実データ max26page (=208sector)
	addr = 0x4000; 
	ptr = (word*)mptr + 512 ;
	for (i = 0 ; i < Sector ; i++ )  {
	    for (j = 0 ; j < 16 ; j++ )  {		// 8W * 16 = 128W
	    	for (k = 0 ; k < 8 ; k++ )  {		// 8W -> d[8] copy
		    d[k] = *ptr++;
		}
	    	sprintf(txbuf, "%04X  %04X %04X %04X %04X %04X %04X %04X %04X \r", addr,
			d[0],d[1],d[2],d[3],d[4],d[5],d[6],d[7] );
		write(rs232_fd, txbuf, strlen(txbuf) ) ;        // error: r = -1
		printf( txbuf ) ;  printf( "\n" ) ;
		addr += 8 ;
		//delayms( 10) ;		// PIC側が遅いのでゆっくり送るため
	    }
	    //printf( "\n" ) ;

	    // PIC からのRDY（sector write）を待つ
	    pipe_req = 1 ;
	    pipe_read_line() ;				// pipebuf: RDY\r\n
	    pipe_req = 0 ;
	    if (strncmp(pipebuf, "RDY", 3) != 0 )  return 9;
	    // RDY  -> next
	}

	free(mptr);        // 用済み free malloc area
	return 0;
}

// -----------------------------------------------------------------
//	HEX file を読み込んで メモリ上のimage に格納する.
//       & malloc buff(image) 有効セクタ数分　hexwork.txt　に書き出す
// -----------------------------------------------------------------
short	read_hexfile(char *path) {
	FILE *ifp, *ofp;
	char *rc;
	uchar *wkptr, *sec_ptr;	// mptr の work ptr
	int  i, j, k, flg ;
	char txbuf[80];		// 出力編集buf
	word addr, d[8], *ptr ;	 

	ifp = fopen(path,"r");
	if(ifp==NULL) 	{
	     printf("Fatal Error!  can't open file:%s\n",path);
	     return 9;
        }  

	ofp = fopen("hexwork.txt","w");
	if(ofp==NULL) 	{
	     printf("Fatal Error!  can't open file:hexwork.txt \n");
	     return 9;
        }  

//  malloc buff GET
	mptr = (uchar *)malloc(54272);		// PIC24FJ64GA02 (208+4)*256=54272
	if (mptr == NULL) {
	     printf("Fatal Error!  can't malloc \n");
	     return 9;
	}
//  malloc buff を全て 00 でクリアする。
	wkptr = mptr;	// mptr save
	for (i = 0; i < 54272; i++) {
		*wkptr = 0; wkptr++; 
	}
	CONFIG2=0; CONFIG1=0;

//  malloc buff にwrite image を格納
	while(1) {
		rc = fgets(linebuf, 80, ifp);
		if(rc == NULL)  { break; }
		read_ihex(); 	          // 1 line 処理
	}

//  malloc buff(image) を調べ （但しmptr+1024以降にoffsetされたAP）全て0x00以外 
//               のセクタ(256byte)の有効セクタ数を求める。
	Sector = 0 ;
	sec_ptr = mptr + 1024 ;				// 調査開始ptr
	while(1) {
		wkptr = sec_ptr + (Sector * 256) ;
		flg = 0 ;
		for ( i = 0 ; i < 256 ; i++ )  {
			if (*wkptr != 0 )  {flg = 1 ; break; } 
			else wkptr++ ;
		}
		if (flg == 1)  {
			Sector++; 		// 次セクタ調査の準備
			continue;
		}
		else  	break; 			// all=0 検出 おわり
	}

printf("Sector=%d \n", Sector);
printf("Config2=%04x  Config1=%04x \n\n", CONFIG2,CONFIG1);

//  malloc buff(image) 有効セクタ数分　hexwork.txt　に書き出す
//    malloc buff(0100-01ff)　２セクタ　AIVT(代替ベクタテーブル）
	addr = 0x0100; 
	ptr = (word*)mptr + 0x0100 ;
	for (i = 0 ; i < 2 ; i++ )  {
	    for (j = 0 ; j < 16 ; j++ )  {		// 8W * 16 = 128W
	    	for (k = 0 ; k < 8 ; k++ )  {		// 8W -> d[8] copy
		    d[k] = *ptr++;
		}
	    	sprintf(txbuf, "%04X  %04X %04X %04X %04X %04X %04X %04X %04X \r", addr,
			d[0],d[1],d[2],d[3],d[4],d[5],d[6],d[7] );
		fputs(txbuf, ofp);          // txbuf -> hexwork.txt 1line Write
		addr +=  8 ;
	    }
	    //delayms( 1 ) ;				// Flash Write Wait
	    fputs("\r", ofp);          
	}

//  malloc buff(4000-xxxx)  セクタ数分繰り返す。　実データ
	addr = 0x4000; 
	ptr = (word*)mptr + 512 ;
	for (i = 0 ; i < Sector ; i++ )  {
	    for (j = 0 ; j < 16 ; j++ )  {		// 8W * 16 = 128W
	    	for (k = 0 ; k < 8 ; k++ )  {		// 8W -> d[8] copy
		    d[k] = *ptr++;
		}
	    	sprintf(txbuf, "%04X  %04X %04X %04X %04X %04X %04X %04X %04X \r", addr,
			d[0],d[1],d[2],d[3],d[4],d[5],d[6],d[7] );
		fputs(txbuf, ofp);          // txbuf -> hexwork.txt 1line Write
		addr += 8 ;
	    }
	    //delayms( 1 ) ;				// Flash Write Wait
	    fputs("\r", ofp);          
	}

	//free(mptr);        // まだ使う　最後に
	fclose(ifp);
	fclose(ofp);
	return 0;
}

// ---------------------------------------------------------------------
//   read_ihex()     hex file の　1line を処理する。		
//      :1000100084C083C082C081C080C07FC07EC07DC0DC
//         データ長 アドレス レコードタイプ   データ  チェックサム.
//      1line buff  from  xxx.hex
// ---------------------------------------------------------------------
short	read_ihex(void) {	

	if(linebuf[0] == ':') {
		read_hex_string();  	// 16進で書かれた文字列をバイナリーに変換し databuf[40]に格納
		parse_ihex();		// record type =00 ならimage 作成
	}
	return 0;
}

// -----------------------------------------------------------------------
//	linebuf(文字列) ->バイナリーに変換-> databuf[40]に格納
//
// -----------------------------------------------------------------------
void read_hex_string(void) {
	char *ptr;
	short idx = 0;
	short rc;

	ptr = linebuf + 1 ;		// 1line buff[1] from  xxx.hex  先頭の: の次を指す
	while(*ptr) {
		rc = read_hex_byte(ptr);  ptr += 2;    // pointer２文字分進める  rc=int
		if (rc<0)  return ;
		databuf[idx++]=(uchar)rc;	 	// binaryデータで格納 max 40byte 
	}
}

// ----------------------------------------------------------------------
//	INTEL HEXレコード(１行:バイナリー化済み) を解釈する.  1行分
//	  record-type=00  malloc buf 書き込み 1line分
// ----------------------------------------------------------------------
short	parse_ihex(void ) {
	uchar record_type;

	record_type = databuf[3];
	switch(record_type) {
		case 0x00:		//データフィールドは書き込まれるべきデータである。
			image_write();	//  malloc buf 書き込み 1line分
			break;		
		case 0x01:		//エンドレコード:
					//HEXファイルの終了を示す.付加情報無し.
			break;

		case 0x02:		//セグメントアドレスレコード:
					//データフィールドにはセグメントアドレスが入る。 たとえば、
					//:02000002E0100C
			break;
		case 0x03:		//スタートセグメントアドレスレコード.
			break;
		case 0x04:		//拡張リニアアドレスレコード.
					//このレコードでは32ビットアドレスのうち上位16ビット
					//（ビット32?ビット16の）を与える。
			break;
		case 0x05:		// スタートリニアアドレス：
					// 例えば    :04000005FF000123D4
					// となっていれば、FF000123h番地がスタートアドレスになる。
			break;
	}
	return (short)record_type;
}

// ----------------------------------------------------------------------
//	INTEL HEXレコード 形式01 (バイナリー化済み) をmalloc buff へ格納する.  1行分
//             databuf[40] static に格納済み
//             databuf[0]=size databuf[1:2]=addr databuf[4]=first data databuf[5]=next
// ----------------------------------------------------------------------
void	image_write( void ) {		//  malloc buf 書き込み 1line分
	uchar *wkptr, i;
	word addr;

	addr = (word)(databuf[1]*256 + databuf[2]);
//  もし　addr=0x57F8(config2), 0x57FC(config1)　なら別処理(config) 
	if      (addr == 0x57F8) { CONFIG2 = databuf[4] + databuf[5]*256 ; }
	else if (addr == 0x57FC) { CONFIG1 = databuf[4] + databuf[5]*256 ;}

//  もし　addr < 0x0400 ならそのまま addr >= 0x8000　ならoffset してmalloc buff へ格納する 
	else {
		if      (addr < 0x0400)  wkptr = mptr + addr ;	// そのまま malloc buf に
		else if (addr >= 0x8000) wkptr = mptr + (addr-0x8000)+1024;	// 1024byte offset
		else  return;					// error 何もしない

		for (i = 0; i < databuf[0]; i++)  {		// size分 repeat
			*wkptr = databuf[4+i];  wkptr++;
		}
	}
}

// ----------------------------------------------------------
//	from file  一行入力ユーティリティ
// ----------------------------------------------------------
short getln(char *buf, FILE *fp) {
	short   c,  l;
	l=0;
	while(1) {
		c=getc(fp);
		if(c==EOF)  {
			*buf = 0;
			return(EOF);    // EOF 
		}
		if(c==0x0d) continue;
		if(c==0x0a) {
			*buf = 0;	// EOL
			return(0);	// OK  
		}
		*buf++ = c;l++;
		if(l>=63) {
			*buf = 0;
			return(1);	// Too long line
		}
	}
}
// --------------------------------------------------------
//	16進2桁の文字列をバイナリーに変換する.
// --------------------------------------------------------
short read_hex_byte(char *s) {
	char buf[4];
	int rc = -1;

	buf[0] = s[0];
	buf[1] = s[1];
	buf[2] = 0;
	sscanf(buf,"%x",&rc);
	return (short)rc;    // rc= int
}

// --------------------------------------------
//   HEX文字列(max 4)から word数値に変換
//     return 0xffff: Error 
// --------------------------------------------
word atohex(char *buf ) {
	uchar    in ;
	word	out = 0 ;

	while (1) {
		in = *buf; buf++;
		if (in == 0) break;
		if ((in >= '0') && (in <= '9')) {
		    	out = (out << 4) + (in & 0x0f) ; 
		}
		else if ((in >= 'a') && (in <= 'f')) {		// a=0x61(97)
		    	out = (out << 4) + (in - 87) ; 
		}
		else if ((in >= 'A') && (in <= 'F')) {		// A=0x41(65)
		    	out = (out << 4) + (in - 55) ; 
		}
		else out=0xffff;
	}
	return out;
}

// hex関連 EOF ------------------------------------------------------------------------------

void delayms(short T) {
	usleep(1000 * T);
}
void delayus(short T) {
	usleep( T);
}
void delayns(short T) {
	short i;
	for (i = 0; i < T; i++) { ; }
}

// EOF




